"""
Common utilities for Actuaries who want to use Python
"""

def format_perc(x, dec_places=1):
    """
    format_perc

    x: a number to format as a percentage
    dec_places: an integer for the number of digits past the decimal
    """
    format_str = '{{:,.{}f}}%'.format(dec_places)
    return format_str.format(float(100*x))


"""
This class is used to represent an actuarial rate table
"""


class RateTable:

    def __init__(self, age_start, age_end, step, left):
        """
        This initializes the rates based on a start and end age
        Parameters
        __________
        age_start: the starting age
        age_end: the ending age
        step: the width of each age band
        left: whether or not an age will match the left endpoint or right endpoint
        """
        self.age_start = age_start
        self.age_end = age_end
        self.step = step
        self.ages = list(range(age_start, age_end + 1, step)) # Need to add one to include age_end
        self.left = left
        self.rates = []

    def set_rates(self, rates):
        """
        Set the rates for every age band
        :param rates: a list of rates for the age band
        :return: whether or not the rates were set for the rate table
        """
        if len(rates) == len(self.ages):
            self.rates.extend(rates)
            return True
        else:
            return False

    def get_rate(self, age):
        """
        Get a single rate for a given age
        :param age: the lookup age
        :return: the rate for the lookup age or 0 if the rates are not set for the rate table
        """
        ages = self.ages
        left = self.left
        rates = self.rates
        if len(rates) == 0:
            return 0
        if age <= min(ages):
            age_index = ages.index(min(ages))
        elif age >= max(ages):
            age_index = ages.index(max(ages))
        else:
            greater_than_age = list(map(lambda x: age < x, ages))
            age_index = greater_than_age.index(True)
            if left:
                age_index = age_index - 1
        return rates[age_index]

    def __repr__(self):
        ages = self.ages
        rates = self.rates
        if len(rates) > 0:
            table_rows = ['Age {} has rate {}'.format(age, rate) for age, rate in zip(ages, rates)]
        else:
            table_rows = ['Age {}'.format(age) for age in ages]
        return '\n'.join(table_rows)



import unittest  # similar to testthat library in R
from RateTable import RateTable


class RateTableTest(unittest.TestCase):

    def setUp(self):  # anything done in setup is available for each test.  Tests are isolated from one another.
        self.rate_table = RateTable(65, 75, 5, True)
        self.rate_table.set_rates([1000, 1500, 2000])

    def test_ages(self):  # needs to start with test_
        ages = self.rate_table.ages
        self.assertEqual(ages, list(range(65, 76, 5)))

    def test_setting_rates(self):
        rate_table = self.rate_table
        self.assertEqual(rate_table.get_rate(72), 1500)


if __name__ == '__main__':
    unittest.main()


"""
Module with helper functions to help with linear regression
"""
import pandas as pd
import numpy as np
from scipy import stats
from sklearn import linear_model

from ..utils.formatting import format_perc


def lm(X, y, sample_weight=None):
    """
    lm

    Enhanced sklearn code for running a linear regression including pandas support and p-values
    Returns a linear model with R^2 automatically calculated as the r2 attribute and T-statistics/p-values as result attribute
    X: A DataFrame containing the independent variables
    y: A ndarray containing the outcome variable
    sample_weight: An optional argument containing sample weights for each row of X
    """
    X_cols = X.columns.tolist()
    X = np.array(X)
    y = y.values.reshape(-1, 1)
    model = linear_model.LinearRegression()
    if sample_weight is not None:
        model.fit(X, y, sample_weight=sample_weight)
    else:
        model.fit(X, y)
    params = np.append(model.intercept_, model.coef_)
    predictions = model.predict(X)
    newX = pd.DataFrame({"Constant":np.ones(len(X))}).join(pd.DataFrame(X))
    MSE = (sum((y-predictions)**2))/(len(newX)-len(newX.columns))
    var_b = MSE*(np.linalg.inv(np.dot(newX.T,newX)).diagonal())
    sd_b = np.sqrt(var_b)
    ts_b = params/ sd_b
    p_values =[2*(1-stats.t.cdf(np.abs(i),(len(newX)-1))) for i in ts_b]
    sd_b = np.round(sd_b,3)
    ts_b = np.round(ts_b,3)
    p_values = np.round(p_values,3)
    params = np.round(params,4)
    result = pd.DataFrame([params, sd_b, ts_b, p_values]).transpose()
    result.columns = ['coef','se','t_val','p_val']
    result.index=['intercept'] + X_cols
    model.result = result
    model.r2 = format_perc(model.score(X, y, sample_weight=sample_weight))
    return model


from setuptools import setup

setup(
    name='actuary',
    version='0.1',
    packages=['actuary', 'actuary.utils', 'actuary.regression',
              'Chapter_2_Core_Python.Chapter_2_Case_Study.rate_table'],
    url='https://github.com/validatehealth/actuary',
    license='MIT',
    author='Andrew Webster',
    author_email='andrew.webster@validatehealth.com',
    description='This is an actuarial toolkit as introduced by three Actex webinars on Python for Actuaries'
)
